#ifndef MUBLAS_H
#define MUBLAS_H

#include <cuComplex.h>

// ***** FORMAT *****

struct __align__(8) cudfreal
{
	float x;
	float y;
};

struct __align__(16) cuddreal
{
	double x;
	double y;
};

struct __align__(16) cudfcomplex
{
	cudfreal x;
	cudfreal y;
};

struct cuddcomplex
{
	cuddreal x;
	cuddreal y;
};

// cudfreal ----------------------------------
__host__ __device__ __forceinline__ cudfreal
make_cudfreal 
	(const float h, const float l)
{
	cudfreal v;
	v.x = h;
	v.y = l;
	return v;
}
// cudfreal ----------------------------------

// cudfcomplex ----------------------------------
__host__ __device__ __forceinline__ cudfcomplex
make_cudfcomplex 
	(const cudfreal re, const cudfreal im)
{
	cudfcomplex v;
	v.x = re;
	v.y = im;
	return v;
}

__host__ __device__ __forceinline__ cudfcomplex
make_cudfcomplex
	(const float reh, const float rel, const float imh, const float iml)
{
	cudfcomplex v;
	v.x.x = reh;
	v.x.y = rel;
	v.y.x = imh;
	v.y.y = iml;
	return v;
}
// cudfcomplex ----------------------------------

// cuddreal ----------------------------------
__host__ __device__ __forceinline__ cuddreal
make_cuddreal 
	(const double h, const double l)
{
	cuddreal v;
	v.x = h;
	v.y = l;
	return v;
}

__host__ __device__ __forceinline__ cuddreal
make_cuddreal 
	(const double h)
{
	cuddreal v;
	v.x = h;
	v.y = 0.;
	return v;
}
// cuddreal ----------------------------------

// cuddcomplex ----------------------------------
__host__ __device__ __forceinline__ cuddcomplex
make_cuddcomplex 
	(const cuddreal re, const cuddreal im)
{
	cuddcomplex v;
	v.x = re;
	v.y = im;
	return v;
}

__host__ __device__ __forceinline__ cuddcomplex
make_cuddcomplex 
	(const double re, const double im)
{
	cuddcomplex v;
	v.x.x = re;
	v.x.y = 0.;
	v.y.x = im;
	v.y.y = 0.;
	return v;
}

__host__ __device__ __forceinline__ cuddcomplex
make_cuddcomplex
	(const double reh, const double rel, const double imh, const double iml)
{
	cuddcomplex v;
	v.x.x = reh;
	v.x.y = rel;
	v.y.x = imh;
	v.y.y = iml;
	return v;
}
// cuddcomplex ----------------------------------

struct mublasHandle_t {
	// mublas parameter
	int32_t mublasVersion;	// 00.00.00
	int32_t mublaserror;	// error handler
	int32_t count;			// 
	float *work_addr;
	float *work_addr2;
	int32_t debug;			// 

	// device parameter
	int32_t	devprop_multiProcessorCount;
	int32_t devprop_sharedMemPerMultiprocessor;
	int32_t devprop_regsPerMultiprocessor;
	int32_t devprop_sharedMemPerBlock;
	int32_t devprop_regsPerBlock;
	int32_t devprop_major;
	int32_t driverVersion; 
	int32_t runtimeVersion;
	int32_t cc;

	// architecture parameter
	SIZE_T WrpPerMpPmax;		// Physical Max of Warps per Multiprocessor
	SIZE_T BlkPerMpPmax;		// Physical Max of Blocks per Multiprocessor
	SIZE_T RegUnit;			// Register Allocation Unit
	SIZE_T ShmUnit;			// Shared Memory Allocation Unit
	SIZE_T WrpGran;			// Warp Allocation Granularity
	SIZE_T Cores;			// # of CUDA Cores per Multiprocessor
	SIZE_T DPCores;		// # of DP Cores per Multiprocessor

	// kernel parameters
	cudaFuncAttributes funcatb;
	int32_t kernelnumber;	// Kernel Number
	void (*func_getShmPerBlk)(mublasHandle_t* mh);
	void (*func_getBlkPerGrd)(mublasHandle_t* mh);
	SIZE_T N_T;
	SIZE_T BlkDim;
	float WrpOcp;	// Warp Occupancy
	float BlkOcp;	// Block Occupancy
	float GrdOcp;	// Grid Occupancy
	SIZE_T Tx;
	SIZE_T Ty;
	SIZE_T m;
	SIZE_T n;
	SIZE_T k;
	SIZE_T ShmPerBlk;
	SIZE_T BlkPerDev;
	SIZE_T BlkPerMpMax;
	SIZE_T N_one;

	// tuning parameter
	float WrpOcpMin;
	float BlkOcpMin;
	float GrdOcpMin;
	SIZE_T TxMin;
	SIZE_T TxMax;
	SIZE_T TxStep;
	SIZE_T TxStatic;
	SIZE_T TyStatic;
	char TxPrefer;
	char TyPrefer;
	SIZE_T ThMin;
	SIZE_T ThMax;
	float TyPerTxMax;

};

void mublasCreate (mublasHandle_t*);
void mublasDestroy (mublasHandle_t*);

// GEMM
extern int32_t mublasSgemm (mublasHandle_t*, const char, const char, const SIZE_T, const SIZE_T, const SIZE_T, const float*, const float*, const SIZE_T, const float*, const SIZE_T, const float*, float*, const SIZE_T);
extern int32_t mublasDgemm (mublasHandle_t*, const char, const char, const SIZE_T, const SIZE_T, const SIZE_T, const double*, const double*, const SIZE_T, const double*, const SIZE_T, const double*, double*, const SIZE_T);
extern int32_t mublasCgemm (mublasHandle_t*, const char, const char, const SIZE_T, const SIZE_T, const SIZE_T, const cuFloatComplex*, const cuFloatComplex*, const SIZE_T, const cuFloatComplex*, const SIZE_T, const cuFloatComplex*, cuFloatComplex*, const SIZE_T);
extern int32_t mublasZgemm (mublasHandle_t*, const char, const char, const SIZE_T, const SIZE_T, const SIZE_T, const cuDoubleComplex*, const cuDoubleComplex*, const SIZE_T, const cuDoubleComplex*, const SIZE_T, const cuDoubleComplex*, cuDoubleComplex*, const SIZE_T);
extern int32_t mublasWgemm (mublasHandle_t*, const char, const char, const SIZE_T, const SIZE_T, const SIZE_T, const cuddreal*, const cuddreal*, const SIZE_T, const cuddreal*, const SIZE_T, const cuddreal*, cuddreal*, const SIZE_T);
extern int32_t mublasUgemm (mublasHandle_t*, const char, const char, const SIZE_T, const SIZE_T, const SIZE_T, const cuddcomplex*, const cuddcomplex*, const SIZE_T, const cuddcomplex*, const SIZE_T, const cuddcomplex*, cuddcomplex*, const SIZE_T);
extern int32_t mublasBgemm (mublasHandle_t*, const char, const char, const SIZE_T, const SIZE_T, const SIZE_T, const cudfreal*, const cudfreal*, const SIZE_T, const cudfreal*, const SIZE_T, const cudfreal*, cudfreal*, const SIZE_T);
extern int32_t mublasVgemm (mublasHandle_t*, const char, const char, const SIZE_T, const SIZE_T, const SIZE_T, const cudfcomplex*, const cudfcomplex*, const SIZE_T, const cudfcomplex*, const SIZE_T, const cudfcomplex*, cudfcomplex*, const SIZE_T);

// GEMV
extern int32_t mublasSgemv (mublasHandle_t*, const char, const SIZE_T, const SIZE_T, const float*, const float*, const SIZE_T, const float*, const SIZE_T, const float*, float*, const SIZE_T);
extern int32_t mublasDgemv (mublasHandle_t*, const char, const SIZE_T, const SIZE_T, const double*, const double*, const SIZE_T, const double*, const SIZE_T, const double*, double*, const SIZE_T);
extern int32_t mublasCgemv (mublasHandle_t*, const char, const SIZE_T, const SIZE_T, const cuFloatComplex*, const cuFloatComplex*, const SIZE_T, const cuFloatComplex*, const SIZE_T, const cuFloatComplex*, cuFloatComplex*, const SIZE_T);
extern int32_t mublasZgemv (mublasHandle_t*, const char, const SIZE_T, const SIZE_T, const cuDoubleComplex*, const cuDoubleComplex*, const SIZE_T, const cuDoubleComplex*, const SIZE_T, const cuDoubleComplex*, cuDoubleComplex*, const SIZE_T);
extern int32_t mublasWgemv (mublasHandle_t*, const char, const SIZE_T, const SIZE_T, const cuddreal*, const cuddreal*, const SIZE_T, const cuddreal*, const SIZE_T, const cuddreal*, cuddreal*, const SIZE_T);
extern int32_t mublasUgemv (mublasHandle_t*, const char, const SIZE_T, const SIZE_T, const cuddcomplex*, const cuddcomplex*, const SIZE_T, const cuddcomplex*, const SIZE_T, const cuddcomplex*, cuddcomplex*, const SIZE_T);
extern int32_t mublasBgemv (mublasHandle_t*, const char, const SIZE_T, const SIZE_T, const cudfreal*, const cudfreal*, const SIZE_T, const cudfreal*, const SIZE_T, const cudfreal*, cudfreal*, const SIZE_T);
extern int32_t mublasVgemv (mublasHandle_t*, const char, const SIZE_T, const SIZE_T, const cudfcomplex*, const cudfcomplex*, const SIZE_T, const cudfcomplex*, const SIZE_T, const cudfcomplex*, cudfcomplex*, const SIZE_T);

// TRMV
extern int32_t mublasStrmv (mublasHandle_t*, const char, const char, const char, const SIZE_T, const float*, const SIZE_T, float*, const SIZE_T);
extern int32_t mublasDtrmv (mublasHandle_t*, const char, const char, const char, const SIZE_T, const double*, const SIZE_T, double*, const SIZE_T);
extern int32_t mublasCtrmv (mublasHandle_t*, const char, const char, const char, const SIZE_T, const cuFloatComplex*, const SIZE_T, cuFloatComplex*, const SIZE_T);
extern int32_t mublasZtrmv (mublasHandle_t*, const char, const char, const char, const SIZE_T, const cuDoubleComplex*, const SIZE_T, cuDoubleComplex*, const SIZE_T);
extern int32_t mublasWtrmv (mublasHandle_t*, const char, const char, const char, const SIZE_T, const cuddreal*, const SIZE_T, cuddreal*, const SIZE_T);
extern int32_t mublasUtrmv (mublasHandle_t*, const char, const char, const char, const SIZE_T, const cuddcomplex*, const SIZE_T, cuddcomplex*, const SIZE_T);
extern int32_t mublasBtrmv (mublasHandle_t*, const char, const char, const char, const SIZE_T, const cudfreal*, const SIZE_T, cudfreal*, const SIZE_T);
extern int32_t mublasVtrmv (mublasHandle_t*, const char, const char, const char, const SIZE_T, const cudfcomplex*, const SIZE_T, cudfcomplex*, const SIZE_T);

// AXPY
extern int32_t mublasSaxpy (mublasHandle_t*, const SIZE_T, const float*, const float*, const SIZE_T, float*, const SIZE_T);
extern int32_t mublasDaxpy (mublasHandle_t*, const SIZE_T, const double*, const double*, const SIZE_T, double*, const SIZE_T);
extern int32_t mublasCaxpy (mublasHandle_t*, const SIZE_T, const cuFloatComplex*, const cuFloatComplex*, const SIZE_T, cuFloatComplex*, const SIZE_T);
extern int32_t mublasZaxpy (mublasHandle_t*, const SIZE_T, const cuDoubleComplex*, const cuDoubleComplex*, const SIZE_T, cuDoubleComplex*, const SIZE_T);
extern int32_t mublasWaxpy (mublasHandle_t*, const SIZE_T, const cuddreal*, const cuddreal*, const SIZE_T, cuddreal*, const SIZE_T);
extern int32_t mublasUaxpy (mublasHandle_t*, const SIZE_T, const cuddcomplex*, const cuddcomplex*, const SIZE_T, cuddcomplex*, const SIZE_T);
extern int32_t mublasBaxpy (mublasHandle_t*, const SIZE_T, const cudfreal*, const cudfreal*, const SIZE_T, cudfreal*, const SIZE_T);
extern int32_t mublasVaxpy (mublasHandle_t*, const SIZE_T, const cudfcomplex*, const cudfcomplex*, const SIZE_T, cudfcomplex*, const SIZE_T);

// DOT
extern int32_t mublasSdot (mublasHandle_t*, const SIZE_T, const float*, const SIZE_T, const float*, const SIZE_T, float*);
extern int32_t mublasDdot (mublasHandle_t*, const SIZE_T, const double*, const SIZE_T, const double*, const SIZE_T, double*);
extern int32_t mublasCdot (mublasHandle_t*, const SIZE_T, const cuFloatComplex*, const SIZE_T, const cuFloatComplex*, const SIZE_T, cuFloatComplex*);
extern int32_t mublasZdot (mublasHandle_t*, const SIZE_T, const cuDoubleComplex*, const SIZE_T, const cuDoubleComplex*, const SIZE_T, cuDoubleComplex*);
extern int32_t mublasWdot (mublasHandle_t*, const SIZE_T, const cuddreal*, const SIZE_T, const cuddreal*, const SIZE_T, cuddreal*);
extern int32_t mublasUdot (mublasHandle_t*, const SIZE_T, const cuddcomplex*, const SIZE_T, const cuddcomplex*, const SIZE_T, cuddcomplex*);
extern int32_t mublasBdot (mublasHandle_t*, const SIZE_T, const cudfreal*, const SIZE_T, const cudfreal*, const SIZE_T, cudfreal*);
extern int32_t mublasVdot (mublasHandle_t*, const SIZE_T, const cudfcomplex*, const SIZE_T, const cudfcomplex*, const SIZE_T, cudfcomplex*);

// COPY
extern int32_t mublasScopy (mublasHandle_t*, const SIZE_T, const float*, const SIZE_T, float*, const SIZE_T);
extern int32_t mublasDcopy (mublasHandle_t*, const SIZE_T, const double*, const SIZE_T, double*, const SIZE_T);
extern int32_t mublasCcopy (mublasHandle_t*, const SIZE_T, const cuFloatComplex*, const SIZE_T, cuFloatComplex*, const SIZE_T);
extern int32_t mublasZcopy (mublasHandle_t*, const SIZE_T, const cuDoubleComplex*, const SIZE_T, cuDoubleComplex*, const SIZE_T);
extern int32_t mublasWcopy (mublasHandle_t*, const SIZE_T, const cuddreal*, const SIZE_T, cuddreal*, const SIZE_T);
extern int32_t mublasUcopy (mublasHandle_t*, const SIZE_T, const cuddcomplex*, const SIZE_T, cuddcomplex*, const SIZE_T);
extern int32_t mublasBcopy (mublasHandle_t*, const SIZE_T, const cudfreal*, const SIZE_T, cudfreal*, const SIZE_T);
extern int32_t mublasVcopy (mublasHandle_t*, const SIZE_T, const cudfcomplex*, const SIZE_T, cudfcomplex*, const SIZE_T);

// SCAL
extern int32_t mublasSscal (mublasHandle_t*, const SIZE_T, const float*, float*, const SIZE_T);
extern int32_t mublasDscal (mublasHandle_t*, const SIZE_T, const double*, double*, const SIZE_T);
extern int32_t mublasCscal (mublasHandle_t*, const SIZE_T, const cuFloatComplex*, cuFloatComplex*, const SIZE_T);
extern int32_t mublasZscal (mublasHandle_t*, const SIZE_T, const cuDoubleComplex*, cuDoubleComplex*, const SIZE_T);
extern int32_t mublasWscal (mublasHandle_t*, const SIZE_T, const cuddreal*, cuddreal*, const SIZE_T);
extern int32_t mublasUscal (mublasHandle_t*, const SIZE_T, const cuddcomplex*, cuddcomplex*, const SIZE_T);
extern int32_t mublasBscal (mublasHandle_t*, const SIZE_T, const cudfreal*, cudfreal*, const SIZE_T);
extern int32_t mublasVscal (mublasHandle_t*, const SIZE_T, const cudfcomplex*, cudfcomplex*, const SIZE_T);

extern int32_t mublasScsrmv (mublasHandle_t*, const char, const SIZE_T, const SIZE_T, const SIZE_T, const float*, const float*, const int32_t*, const int32_t*, const float*, const float*, float*);
extern int32_t mublasDcsrmv (mublasHandle_t*, const char, const SIZE_T, const SIZE_T, const SIZE_T, const double*, const double*, const int32_t*, const int32_t*, const double*, const double*, double*);
extern int32_t mublasCcsrmv (mublasHandle_t*, const char, const SIZE_T, const SIZE_T, const SIZE_T, const cuFloatComplex*, const cuFloatComplex*, const int32_t*, const int32_t*, const cuFloatComplex*, const cuFloatComplex*, cuFloatComplex*);
extern int32_t mublasZcsrmv (mublasHandle_t*, const char, const SIZE_T, const SIZE_T, const SIZE_T, const cuDoubleComplex*, const cuDoubleComplex*, const int32_t*, const int32_t*, const cuDoubleComplex*, const cuDoubleComplex*, cuDoubleComplex*);
extern int32_t mublasWcsrmv (mublasHandle_t*, const char, const SIZE_T, const SIZE_T, const SIZE_T, const cuddreal*, const cuddreal*, const int32_t*, const int32_t*, const cuddreal*, const cuddreal*, cuddreal*);
extern int32_t mublasUcsrmv (mublasHandle_t*, const char, const SIZE_T, const SIZE_T, const SIZE_T, const cuddcomplex*, const cuddcomplex*, const int32_t*, const int32_t*, const cuddcomplex*, const cuddcomplex*, cuddcomplex*);
extern int32_t mublasBcsrmv (mublasHandle_t*, const char, const SIZE_T, const SIZE_T, const SIZE_T, const cudfreal*, const cudfreal*, const int32_t*, const int32_t*, const cudfreal*, const cudfreal*, cudfreal*);
extern int32_t mublasVcsrmv (mublasHandle_t*, const char, const SIZE_T, const SIZE_T, const SIZE_T, const cudfcomplex*, const cudfcomplex*, const int32_t*, const int32_t*, const cudfcomplex*, const cudfcomplex*, cudfcomplex*);

#endif
