#include "mublas_devel.h"

void
mublasCreate (mublasHandle_t *mh) {
    int32_t driverVersion = 0;
	int32_t runtimeVersion = 0;
	int32_t devid;
	cudaDriverGetVersion(&driverVersion);
	cudaRuntimeGetVersion(&runtimeVersion);

	// Device Properties
	cudaDeviceProp devprop;  
	cudaGetDevice (&devid);
	cudaGetDeviceProperties (&devprop, devid);
	mh->devprop_multiProcessorCount = devprop.multiProcessorCount;
	mh->devprop_sharedMemPerMultiprocessor = devprop.sharedMemPerMultiprocessor;
	mh->devprop_regsPerMultiprocessor = devprop.regsPerMultiprocessor;
	mh->devprop_sharedMemPerBlock = devprop.sharedMemPerBlock;
	mh->devprop_regsPerBlock = devprop.regsPerBlock;
	mh->devprop_major = devprop.major;
	mh->cc = devprop.major * 100 + devprop.minor * 10;
	mh->driverVersion = driverVersion;
	mh->runtimeVersion = runtimeVersion;

	mh->mublasVersion = MUBLAS_VERSION;
	mh->mublaserror = 1111; // set 1111 when mublasCreate called
	mh->kernelnumber = 0;
	mh->count = 0;
	mh->debug = 0;

	// tuning parameter, default value
	#include "mublas_all.incl"

	// default value (architecture dependent)
	switch (mh->cc) {
		case 200:
			mh->WrpPerMpPmax = 48;
			mh->BlkPerMpPmax = 8;
			mh->RegUnit = 64;
			mh->ShmUnit = 128;
			mh->WrpGran = 2;
			mh->Cores = 32;
			mh->DPCores = 16;
			break;
		case 210:
			mh->WrpPerMpPmax = 48;
			mh->BlkPerMpPmax = 8;
			mh->RegUnit = 64;
			mh->ShmUnit = 128;
			mh->WrpGran = 2;
			mh->Cores = 48;
			mh->DPCores = 6;
			break;
		case 300:
			mh->WrpPerMpPmax = 64;
			mh->BlkPerMpPmax = 16;
			mh->RegUnit = 256;
			mh->ShmUnit = 256;
			mh->WrpGran = 4;
			mh->Cores = 192;
			mh->DPCores = 64;
			break;
		case 320:
			mh->WrpPerMpPmax = 64;
			mh->BlkPerMpPmax = 16;
			mh->RegUnit = 256;
			mh->ShmUnit = 256;
			mh->WrpGran = 4;
			mh->Cores = 192;
			mh->DPCores = 64;
			break;
		case 350:
			mh->WrpPerMpPmax = 64;
			mh->BlkPerMpPmax = 16;
			mh->RegUnit = 256;
			mh->ShmUnit = 256;
			mh->WrpGran = 4;
			mh->Cores = 192;
			mh->DPCores = 64;
			break;
		case 370:
			mh->WrpPerMpPmax = 64;
			mh->BlkPerMpPmax = 16;
			mh->RegUnit = 256;
			mh->ShmUnit = 256;
			mh->WrpGran = 4;
			mh->Cores = 192;
			mh->DPCores = 64;
			break;
		case 500:
			mh->WrpPerMpPmax = 64;
			mh->BlkPerMpPmax = 32;
			mh->RegUnit = 256;
			mh->ShmUnit = 256;
			mh->WrpGran = 4;
			mh->Cores = 128;
			mh->DPCores = 4;
			break;
		case 520:
			mh->WrpPerMpPmax = 64;
			mh->BlkPerMpPmax = 32;
			mh->RegUnit = 256;
			mh->ShmUnit = 256;
			mh->WrpGran = 4;
			mh->Cores = 128;
			mh->DPCores = 4;
			break;
		default: // non-supported architecture
			mh->mublaserror = 9999;
			break;
	}

	// work space allocation
	float *w;
	cudaMalloc ((void **) &w, sqrt((float)devprop.totalGlobalMem/(sizeof(double)*4))*sizeof(double)*4);
	mh->work_addr = w;
	float *w2;
	cudaMalloc ((void **) &w2, sqrt((float)devprop.totalGlobalMem/(sizeof(double)*4))*sizeof(double)*4);
	mh->work_addr2 = w2;
}

void
mublasDestroy (mublasHandle_t *mh) {
	cudaFree (mh->work_addr);
	cudaFree (mh->work_addr2);
}

/* ***********************************************
	Occupancy Calculation
   *********************************************** */   

__host__ void GetOcp ( mublasHandle_t *mh) {
	const SIZE_T ShmPerMpPmax = mh->devprop_sharedMemPerMultiprocessor;
	const SIZE_T MpPerDev = mh->devprop_multiProcessorCount;
	const SIZE_T RegPerMp = mh->devprop_regsPerMultiprocessor;
	const SIZE_T RegPerTh = mh->funcatb.numRegs;
	const SIZE_T RegPerWrp = ceilingEx ((float)(RegPerTh * WrpSize), mh->RegUnit); // Usage of registers per warp
	const SIZE_T WrpPerMpRmax = min (floorEx ((float)RegPerMp / (float)RegPerWrp, mh->WrpGran), mh->WrpPerMpPmax); // Max warps per multiprocessor
	const SIZE_T WrpPerBlk = (SIZE_T) ceil ((float)(mh->Tx * mh->Ty) / (float)WrpSize); // Warps per block
	const SIZE_T BlkPerMpRmax = (SIZE_T) floor ((float)WrpPerMpRmax / (float)WrpPerBlk); // Blocks per multiprocessor (limited by register)
	SIZE_T BlkPerMpSmax; // Blocks per multiprocessor (limited by shared memory)
	if (mh->ShmPerBlk == 0) {
		mh->BlkPerMpMax = min (BlkPerMpRmax, mh->BlkPerMpPmax); // Blocks per multiprocessor
	} else {
		BlkPerMpSmax = (SIZE_T) floor ((float)ShmPerMpPmax / (float)mh->ShmPerBlk);
		mh->BlkPerMpMax = min (min (BlkPerMpRmax, BlkPerMpSmax), mh->BlkPerMpPmax); // Blocks per multiprocessor
	}
	// Block-Occupancy
	mh->BlkOcp = (float)mh->BlkPerMpMax / (float)mh->BlkPerMpPmax;
	const SIZE_T BlkPerDevMax = mh->BlkPerMpMax * MpPerDev;
	// Grid-Occupancy
	mh->GrdOcp = (float)mh->BlkPerDev / (float)ceilingEx ((float)mh->BlkPerDev, BlkPerDevMax);
	SIZE_T WrpPerMpMax = WrpPerBlk * mh->BlkPerMpMax; // Warps per multiprocessor
	// Warp-Occupancy
	mh->WrpOcp = (float)WrpPerMpMax / (float)mh->WrpPerMpPmax; // Warp Occupancy
}


__host__ void determine_best_thread_block_size_2d (mublasHandle_t *mh) {
	SIZE_T Tx, TxOptimal, TxStart, TxStop, TxStep, TxMin, TxMax;
	SIZE_T Ty, TyOptimal, TyStart, TyStop, TyStep, TyMin, TyMax;
	float GrdOcpOld = 0., BlkOcpOld = 0., WrpOcpOld = 0.;
	void (*getShmPerBlk)(mublasHandle_t *mh);
	void (*getBlkPerGrd)(mublasHandle_t *mh);
	getShmPerBlk = mh->func_getShmPerBlk;
	getBlkPerGrd = mh->func_getBlkPerGrd;

	// --- initial values -------------------------------------
	if (mh->TxPrefer == 'L') {
		TxStart = mh->TxMax;
		TxStop  = mh->TxMin;
		TxStep  = -mh->TxStep;
	} else {
		TxStart = mh->TxMin;
		TxStop  = mh->TxMax;
		TxStep  = mh->TxStep;
	}
	mh->count = 0;
	// --- Tx ------------------------------------------------
	Tx = TxStart;
	while (Tx) {
		mh->Tx = Tx;
		TyMin = (SIZE_T)ceil (mh->ThMin / (float)mh->Tx);
		TyMax = max (1, (SIZE_T)floor ((float)mh->ThMax / (float)mh->Tx));
		if (mh->TyPerTxMax != 0)
			TyMax = min (TyMax, mh->Tx * mh->TyPerTxMax);
		if (mh->TyPrefer == 'L') {
			TyStart = TyMax;
			TyStop  = TyMin;
			TyStep  = -1;
		} else {
			TyStart = TyMin;
			TyStop  = TyMax;
			TyStep  = 1;
		}
		getBlkPerGrd (mh);
		// --- Ty -----------------------------------------------
		Ty = TyStart;
		while (Ty) {
			mh->Ty = Ty;
			mh->count++;
			getShmPerBlk (mh);
			GetOcp (mh);
			if ( mh->GrdOcp > min (GrdOcpOld, mh->GrdOcpMin) &&
				 mh->BlkOcp > min (BlkOcpOld, mh->BlkOcpMin) &&
				 mh->WrpOcp > min (WrpOcpOld, mh->WrpOcpMin)) {
				GrdOcpOld = mh->GrdOcp;
				BlkOcpOld = mh->BlkOcp;
				WrpOcpOld = mh->WrpOcp;
				TxOptimal = Tx;
				TyOptimal = Ty;
			}
			if      (mh->TyPrefer == 'L' && Ty <= TyStop) break;
			else if (mh->TyPrefer == 'S' && Ty >= TyStop) break;
			Ty += TyStep;
		// --- Ty -----------------------------------------------
		}
		if      (mh->TxPrefer == 'L' && Tx <= TxStop) break;
		else if (mh->TxPrefer == 'S' && Tx >= TxStop) break;
		Tx += TxStep;
	}
	// --- Tx ------------------------------------------------
	mh->Tx = TxOptimal;
	mh->Ty = TyOptimal;
}
		
__host__ void determine_best_thread_block_size_1d (mublasHandle_t *mh) {
	SIZE_T Tx, TxOptimal, TxStart, TxStop, TxStep, TxMin, TxMax;
	float GrdOcpOld = 0., BlkOcpOld = 0., WrpOcpOld = 0.;
	void (*getShmPerBlk)(mublasHandle_t *mh);
	void (*getBlkPerGrd)(mublasHandle_t *mh);
	getShmPerBlk = mh->func_getShmPerBlk;
	getBlkPerGrd = mh->func_getBlkPerGrd;
	// --- initial values -------------------------------------
	if (mh->TxPrefer == 'L') {
		TxStart = min (mh->ThMax, mh->TxMax);
		TxStop  = max (mh->ThMin, mh->TxMin);
		TxStep  = -mh->TxStep;
	} else if (mh->TxPrefer == 'S') {
		TxStart = max (mh->ThMin, mh->TxMin);
		TxStop  = min (mh->ThMax, mh->TxMax);
		TxStep  = mh->TxStep;
	} 
	mh->count = 0;
	mh->Ty = 1;
	// --- Tx ------------------------------------------------
	Tx = TxStart;
	while (Tx) {
		mh->Tx = Tx;
		mh->count++;
		getBlkPerGrd (mh);
		getShmPerBlk (mh);
		GetOcp (mh);
		if ( mh->GrdOcp > min (GrdOcpOld, mh->GrdOcpMin) &&
			 mh->BlkOcp > min (BlkOcpOld, mh->BlkOcpMin) &&
			 mh->WrpOcp > min (WrpOcpOld, mh->WrpOcpMin)) {
			GrdOcpOld = mh->GrdOcp;
			BlkOcpOld = mh->BlkOcp;
			WrpOcpOld = mh->WrpOcp;
			TxOptimal = Tx;
		}
		if      (mh->TxPrefer == 'L' && Tx <= TxStop) break;
		else if (mh->TxPrefer == 'S' && Tx >= TxStop) break;
		Tx += TxStep;
	}
	// --- Tx ------------------------------------------------
	mh->Tx = TxOptimal;
	mh->Ty = 1;
}

// phase 2: determining optimal thread-block size
__host__ void kernel_configuration (mublasHandle_t *mh) {
	void (*getShmPerBlk)(mublasHandle_t *mh);
	void (*getBlkPerGrd)(mublasHandle_t *mh);
	getShmPerBlk = mh->func_getShmPerBlk;
	getBlkPerGrd = mh->func_getBlkPerGrd;

	// --- Initial Tx & Ty -------------------------------
	// 1. TxStep (0=auto)
	if (mh->TxStep == 0) mh->TxStep = 128 / (mh->N_one * mh->N_T); // 128bytes/16bytes(vec)=8thread
	// 2. ThMin (0=auto)
	if (mh->ThMin == 0) mh->ThMin = mh->TxStep;
	// 3. ThMax
	mh->ThMax = mh->funcatb.maxThreadsPerBlock;
	// 4. TxMin
	mh->TxMin = mh->TxStep;
	// 5. TxMax
	mh->TxMax = mh->ThMax;
	// ---------------------------------------------------

	if (mh->TxStatic == 0) { 
		if      (mh->BlkDim == 2) determine_best_thread_block_size_2d (mh);
		else if (mh->BlkDim == 1) determine_best_thread_block_size_1d (mh);
		else printf ("mublas error: unsupported BlkDim\n");
	} else { // Kernel is launched with specified Tx and Ty (static mode)
		mh->Tx = mh->TxStatic;
		mh->Ty = mh->TyStatic;
		if (mh->Tx * mh->Ty > mh->funcatb.maxThreadsPerBlock) {
			printf ("mublas error: too many resource requested\n");
			exit (1);
		}
	}
	getBlkPerGrd (mh);
	getShmPerBlk (mh);
	GetOcp (mh);
}

