/* ***********************************************
 
	GEMV for CUDA 

   *********************************************** */   

#include "mublas_devel.h"
#include "mublas_gemv_n_kernel.cu"
#include "mublas_gemv_t_kernel.cu"

/* ***********************************************
	Host Interface
   *********************************************** */   

__host__ int32_t
MUBLAS_GEMV (
	mublasHandle_t* mh,	
	const char trans,
	const SIZE_T m,
	const SIZE_T n,
	const TYPE0 *alpha,
	const TYPE0 *a,
	const SIZE_T lda,
	const TYPE0 *x,
	const SIZE_T incx,
	const TYPE0 *beta,
	TYPE0 *y,
	const SIZE_T incy
) {
	SIZE_T kernel = 0;
	SIZE_T use_vector = 0;
	dim3 threads;
	dim3 grid;

	// ---------------------------------------------
	// 1. Shortcut
    if (m == 0 || n == 0 || (*alpha == ZERO0 && *beta == ONE0 )) 
		return 0;
	#if defined (CUBLAS_SCAL)
	if (*alpha == ZERO0) {
		cublasHandle_t ch;
		cublasCreate (&ch);
		cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
		CUBLAS_SCAL (ch, m, beta, y, incy);
		return 0;
	}
	#endif
	// ---------------------------------------------

	// ---------------------------------------------
	// 2. Kernel Configuration Phase 0
	mh->N_one = sizeof (TYPE0);
	mh->N_T = 1;
	if      (VEC4) mh->N_T = 4;
	else if (VEC2) mh->N_T = 2;
	if (
		(((long)a%(128/sizeof(TYPE0)))%16 == 0) && 
		(((long)x%(128/sizeof(TYPE0)))%16 == 0) && 
		(((long)y%(128/sizeof(TYPE0)))%16 == 0) &&
		(lda%mh->N_T == 0)
		) {
		use_vector = 1;
	} else {
		use_vector = 0;
	}
	mh->m = m;
	mh->n = n;
	// ---------------------------------------------

	// ---------------------------------------------
	// 3. Kernel Configuration Phase 1
	kernel = 999; // default=CUBLAS;
	if (incx == 1 && incy == 1) {
		if (trans == 'N' || trans == 'n') {
			mh->BlkDim = 2;
			mh->func_getShmPerBlk = &getShmPerBlk_GEMV_N;
			mh->func_getBlkPerGrd = &getBlkPerGrd_GEMV_N;
			#include "mublas_gemv_n.incl"
			if (mh->N_T == 4) { 
				if (use_vector) kernel = 140;
				else			kernel = 141;
			} else if (mh->N_T == 2) { 
				if (use_vector) kernel = 120;
				else 			kernel = 121;
			} else {
								kernel = 110;
			}
		} else if (trans == 'T' || trans == 't') {
			mh->BlkDim = 1;
			mh->func_getShmPerBlk = &getShmPerBlk_GEMV_T;
			mh->func_getBlkPerGrd = &getBlkPerGrd_GEMV_T;
			#include "mublas_gemv_t.incl"
			if (mh->N_T == 4) { 
				if (use_vector) kernel = 240;
				else			kernel = 210;
			} else if (mh->N_T == 2) {
				if (use_vector) kernel = 220;
				else			kernel = 210;
			} else {
								kernel = 210;
			}
		}
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 4. Kernel Configuration Phase 2
	switch (kernel) {
		// N-kernel
		case 140:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_N_DEVICE_v4);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_N_DEVICE_v4, cudaFuncCachePreferShared);
			kernel_configuration (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->BlkPerDev);
			MUBLAS_GEMV_N_DEVICE_v4 <<< grid, threads, mh->ShmPerBlk >>> (m, n, *alpha, a, lda, x, *beta, y);
			break;
		case 141:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_N_DEVICE_s4);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_N_DEVICE_s4, cudaFuncCachePreferShared);
			kernel_configuration (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->BlkPerDev);
			MUBLAS_GEMV_N_DEVICE_s4 <<< grid, threads, mh->ShmPerBlk >>> (m, n, *alpha, a, lda, x, *beta, y);
			break;
		case 120:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_N_DEVICE_v2);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_N_DEVICE_v2, cudaFuncCachePreferShared);
			kernel_configuration (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->BlkPerDev);
			MUBLAS_GEMV_N_DEVICE_v2 <<< grid, threads, mh->ShmPerBlk >>> (m, n, *alpha, a, lda, x, *beta, y);
			break;
		case 121:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_N_DEVICE_s2);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_N_DEVICE_s2, cudaFuncCachePreferShared);
			kernel_configuration (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->BlkPerDev);
			MUBLAS_GEMV_N_DEVICE_s2 <<< grid, threads, mh->ShmPerBlk >>> (m, n, *alpha, a, lda, x, *beta, y);
			break;
		case 110:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_N_DEVICE_s1);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_N_DEVICE_s1, cudaFuncCachePreferShared);
			kernel_configuration (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->BlkPerDev);
			MUBLAS_GEMV_N_DEVICE_s1 <<< grid, threads, mh->ShmPerBlk >>> (m, n, *alpha, a, lda, x, *beta, y);
			break;
		// T-kernel
		case 240:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_T_DEVICE_v4);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_T_DEVICE_v4, cudaFuncCachePreferL1);
			kernel_configuration (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->BlkPerDev);
			if (mh->cc >= 300) MUBLAS_GEMV_T_DEVICE_v4 <<< grid, threads >>> (m, n, *alpha, a, lda, x, *beta, y);
			else MUBLAS_GEMV_T_DEVICE_v4 <<< grid, threads, mh->ShmPerBlk >>> (m, n, *alpha, a, lda, x, *beta, y);
			break;
		case 220:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_T_DEVICE_v2);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_T_DEVICE_v2, cudaFuncCachePreferL1);
			kernel_configuration (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->BlkPerDev);
			if (mh->cc >= 300) MUBLAS_GEMV_T_DEVICE_v2 <<< grid, threads >>> (m, n, *alpha, a, lda, x, *beta, y);
			else MUBLAS_GEMV_T_DEVICE_v2 <<< grid, threads, mh->ShmPerBlk >>> (m, n, *alpha, a, lda, x, *beta, y);
			break;
		case 210:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_T_DEVICE_s1);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_T_DEVICE_s1, cudaFuncCachePreferL1);
			kernel_configuration (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->BlkPerDev);
			if (mh->cc >= 300) MUBLAS_GEMV_T_DEVICE_s1 <<< grid, threads >>> (m, n, *alpha, a, lda, x, *beta, y);
			else MUBLAS_GEMV_T_DEVICE_s1 <<< grid, threads, mh->ShmPerBlk >>> (m, n, *alpha, a, lda, x, *beta, y);
			break;
		// CUBLAS
		default:
			#if defined (CUBLAS_GEMV)
			cublasHandle_t ch;
			cublasCreate (&ch);
			cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
			cublasOperation_t tranAcublas = CUBLAS_OP_N;
			if (trans == 'T' || trans == 't') tranAcublas = CUBLAS_OP_T;
			if (trans == 'C' || trans == 'c') tranAcublas = CUBLAS_OP_C;
			CUBLAS_GEMV (ch, tranAcublas, m, n, alpha, a, lda, x, incx, beta, y, incy);
			#else
			printf ("mublas error: not implemented");
			#endif
			break;
	}
	// ---------------------------------------------

	if (mh->kernelnumber < 8000) mh->kernelnumber = kernel;
	return 0;
}


