/* ***********************************************
 
	TRMV-LNN for CUDA 

   *********************************************** */   

#include "mublas_trmv_lnn.def"

/* ***********************************************
    Macro
   *********************************************** */   
#if USE_NC_A_TRMV_LNN
#define RESTRICT_A __restrict__
#define LOAD1_A LOAD1_NC
#define LOAD2_A LOAD2_NC
#define LOAD4_A LOAD4_NC
#else
#define RESTRICT_A
#define LOAD1_A LOAD1___
#define LOAD2_A LOAD2___
#define LOAD4_A LOAD4___
#endif
#if USE_NC_X_TRMV_LNN
#define RESTRICT_X __restrict__
#define LOAD1_X LOAD1_NC
#define LOAD2_X LOAD2_NC
#define LOAD4_X LOAD4_NC
#else
#define RESTRICT_X
#define LOAD1_X LOAD1___
#define LOAD2_X LOAD2___
#define LOAD4_X LOAD4___
#endif

/* ***********************************************
	Functions for determining Tx and Ty
   *********************************************** */   
__host__ __inline__ void getBlkPerGrd_TRMV_LNN ( mublasHandle_t* mh) {
	mh->BlkPerDev = (SIZE_T) ceil ((float)mh->n / (float)(mh->N_T * mh->Tx));
}

__host__ __inline__ void getShmPerBlk_TRMV_LNN ( mublasHandle_t* mh) {
	mh->ShmPerBlk = ceilingEx (2. * (float)(mh->Tx * mh->Ty * mh->N_T * mh->N_one), mh->ShmUnit);
}

/* ***********************************************
	Sub-functions
   *********************************************** */   
__global__ void MUBLAS_TRMV_LNN_DEVICE_final (
	const SIZE_T n,
	const TYPE0 * __restrict__ Wgl,
	const TYPE0 * __restrict__ Wgr,
	TYPE0 * Xg
) {
	const SIZE_T Tx = blockDim.x;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T rowid = (blockIdx.x * Tx + iTx);

	if (rowid < n) 
		Xg[rowid] = ADD1 (Wgl[rowid], Wgr[rowid]);
}

/* ***********************************************
	kernel=s1, uplo=L, trans=N, diag=N
   *********************************************** */   

__global__ void MUBLAS_TRMV_LNN_DEVICE_s1 (
	const SIZE_T n,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_X Xg,
	TYPE0 *Wgl,
	TYPE0 *Wgr
) {
#if VEC1
	const SIZE_T Tx = blockDim.x;
	const SIZE_T Ty = blockDim.y;
	const SIZE_T T_B = Tx * Ty;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iTy = threadIdx.y;
	const SIZE_T iTh = iTy * Tx + iTx;
	const SIZE_T rowid = (blockIdx.x * Tx + iTx);
	SIZE_T i;
	extern __shared__ TYPE1 Xs1[];
	TYPE1 Ar1, Xr1, Tr1l, Tr1r;

	Tr1l = ZERO1;
	Tr1r = ZERO1;

	// main
	for (i = iTy; i < n/2+1+(n%2==0?0:1); i+=Ty) {
		if (i < min(n/2,rowid+1)) {
			LOAD1_X (Xr1, &Xg[i]);
			LOAD1_A (Ar1, &Ag[i * lda + rowid]); 
			Tr1l = MAD1 (Xr1, Ar1, Tr1l);
		}
		if (i >= rowid+1) {
			LOAD1_X (Xr1, &Xg[n-i]);
			LOAD1_A (Ar1, &Ag[(n-i) * lda + n-1-rowid]); 
			Tr1r = MAD1 (Xr1, Ar1, Tr1r);
		}
	}

	// reduction
	Xs1[iTh] = Tr1l;
	Xs1[iTh+T_B] = Tr1r;
	__syncthreads ();
	if (iTy == 0) {
		for (i = 1; i < Ty; i++) {
			Tr1l = ADD1 (Tr1l, Xs1[Tx * i + iTx]);
			Tr1r = ADD1 (Tr1r, Xs1[Tx * i + iTx + T_B]);
		}
		if (rowid < n) {
			Wgl[rowid] = Tr1l;
			Wgr[n-1-rowid] = Tr1r;
		}
	}
#endif
}

/* ***********************************************
	kernel=v2, uplo=L, trans=N, diag=N
   *********************************************** */   

__global__ void MUBLAS_TRMV_LNN_DEVICE_v2 (
	const SIZE_T n,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_X Xg,
	TYPE0 *Wgl,
	TYPE0 *Wgr
) {
#if VEC2
	const SIZE_T N_T = 2;
	const SIZE_T Tx = blockDim.x;
	const SIZE_T Ty = blockDim.y;
	const SIZE_T T_B = Tx * Ty;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iTy = threadIdx.y;
	const SIZE_T iTh = iTy * Tx + iTx;
	const SIZE_T rowid = (blockIdx.x * Tx + iTx) * N_T;
	SIZE_T i;
	extern __shared__ TYPE1 Xs1[];
	TYPE2 Ar2, Tr2l, Tr2r;
	TYPE1 Xr1;

	Tr2l = ZERO2;
	Tr2r = ZERO2;

	// main
	for (i = iTy; i < n/2+1+(n%2==0?0:1); i+=Ty) {
		if (i < min(n/2,rowid+1+1)) {
			LOAD1_X (Xr1, &Xg[i]);
			LOAD2_A (Ar2, &Ag[i * lda + rowid]); 
			if (i >= min(n/2,rowid+1+0)) (TYPE1&)Ar2.x = ZERO1;
			if (i >= min(n/2,rowid+1+1)) (TYPE1&)Ar2.y = ZERO1;
			Tr2l = MAD122 (Xr1, Ar2, Tr2l);
		}
		if (i >= rowid+1) {
			LOAD1_X (Xr1, &Xg[n-i]);
			LOAD2_A (Ar2, &Ag[(n-i) * lda + n-1-1-rowid]); 
			if (i < rowid+1+1) (TYPE1&)Ar2.x = ZERO1;
			if (i < rowid+1+0) (TYPE1&)Ar2.y = ZERO1;
			Tr2r = MAD122 (Xr1, Ar2, Tr2r);
		}
	}

	// reduction
	(TYPE2&)Xs1[iTh*N_T] = Tr2l;
	(TYPE2&)Xs1[(iTh+T_B)*N_T] = Tr2r;
	__syncthreads ();
	if (iTy == 0) {
		for (i = 1; i < Ty; i++) {
			Tr2l = ADD2 (Tr2l, (TYPE2&)Xs1[(Tx * i + iTx      )*N_T]);
			Tr2r = ADD2 (Tr2r, (TYPE2&)Xs1[(Tx * i + iTx + T_B)*N_T]);
		}
		if (rowid < n) {
			(TYPE2&)Wgl[      rowid] = Tr2l;
			(TYPE2&)Wgr[n-1-1-rowid] = Tr2r;
		}
	}

#endif
}

/* ***********************************************
	kernel=v4, uplo=L, trans=N, diag=N
   *********************************************** */   

__global__ void MUBLAS_TRMV_LNN_DEVICE_v4 (
	const SIZE_T n,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_X Xg,
	TYPE0 *Wgl,
	TYPE0 *Wgr
) {
#if VEC4
	const SIZE_T N_T = 4;
	const SIZE_T Tx = blockDim.x;
	const SIZE_T Ty = blockDim.y;
	const SIZE_T T_B = Tx * Ty;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iTy = threadIdx.y;
	const SIZE_T iTh = iTy * Tx + iTx;
	const SIZE_T rowid = (blockIdx.x * Tx + iTx) * N_T;
	SIZE_T i;
	extern __shared__ TYPE1 Xs1[];
	TYPE4 Ar4, Tr4l, Tr4r;
	TYPE1 Xr1;

	Tr4l = ZERO4;
	Tr4r = ZERO4;

	// main
	for (i = iTy; i < n/2+1+(n%2==0?0:1); i+=Ty) {
		if (i < min(n/2,rowid+1+3)) {
			LOAD1_X (Xr1, &Xg[i]);
			LOAD4_A (Ar4, &Ag[i * lda + rowid]); 
			if (i >= min(n/2,rowid+1+0)) Ar4.x = ZERO1;
			if (i >= min(n/2,rowid+1+1)) Ar4.y = ZERO1;
			if (i >= min(n/2,rowid+1+2)) Ar4.z = ZERO1;
			if (i >= min(n/2,rowid+1+3)) Ar4.w = ZERO1;
			Tr4l = MAD144 (Xr1, Ar4, Tr4l);
		}
		if (i >= rowid+1) {
			LOAD1_X (Xr1, &Xg[n-i]);
			LOAD4_A (Ar4, &Ag[(n-i) * lda + n-1-3-rowid]); 
			if (i < rowid+1+3) Ar4.x = ZERO1;
			if (i < rowid+1+2) Ar4.y = ZERO1;
			if (i < rowid+1+1) Ar4.z = ZERO1;
			if (i < rowid+1+0) Ar4.w = ZERO1;
			Tr4r = MAD144 (Xr1, Ar4, Tr4r);
		}
	}

	// reduction
	(TYPE4&)Xs1[iTh*N_T] = Tr4l;
	(TYPE4&)Xs1[(iTh+T_B)*N_T] = Tr4r;
	__syncthreads ();
	if (iTy == 0) {
		for (i = 1; i < Ty; i++) {
			Tr4l = ADD4 (Tr4l, (TYPE4&)Xs1[(Tx * i + iTx      )*N_T]);
			Tr4r = ADD4 (Tr4r, (TYPE4&)Xs1[(Tx * i + iTx + T_B)*N_T]);
		}
		if (rowid < n) {
			(TYPE4&)Wgl[      rowid] = Tr4l;
			(TYPE4&)Wgr[n-1-3-rowid] = Tr4r;
		}
	}


#endif
}

#undef RESTRICT_A
#undef RESTRICT_X
#undef LOAD1_A
#undef LOAD2_A
#undef LOAD4_A
#undef LOAD1_X
#undef LOAD2_X
#undef LOAD4_X
