// ----- testing handler ---------------------------------------------------------------
struct testingHandle_t {
	// device parameter
	cudaDeviceProp devprop;  
	int32_t devno;
	int32_t cc;
	double device_bytes_per_sec;
	double device_flops_per_sec_sp;
	double device_flops_per_sec_dp;

	int32_t Cores;
	int32_t DPCores;

	// testing parameter
	char mode; // c=check, p=perf

	double routine_flops_per_sec_theoretical_peak;
	double routine_flops;
	double routine_bytes;

	size_t word_size;
	size_t TxStatic;
	size_t TyStatic;
	int32_t driverVersion; 
	int32_t runtimeVersion;
	size_t test_points;
	int32_t kernelnumber;
	int32_t cublasatomic;
	int32_t nodisp;

	// BLAS parameters
	// dim
	size_t kl;
	size_t ku;
	size_t dim_start;
	size_t dim_stop;
	size_t dim_step;
	size_t dim_m;
	size_t dim_n;
	size_t dim_k;
	size_t incx;
	size_t incy;
	uint32_t mode_dim_start;
	uint32_t mode_dim_stop;
	uint32_t mode_dim_step;

	size_t const_m;
	size_t const_n;
	size_t const_k;

	// options
	char tranA; // N, T, C
	char tranB; // N, T, C
	char uplo; // U, L
	char diag; // N, U
	char side; // L, R
};

// ----- CUBLAS const value function ---------------------------------------------------------------
inline cublasOperation_t cublas_trans_const_m (char tran) {
	if (tran == 'N' || tran == 'n') return CUBLAS_OP_N;
	if (tran == 'T' || tran == 't') return CUBLAS_OP_T;
	if (tran == 'C' || tran == 'c') return CUBLAS_OP_C;
	return CUBLAS_OP_N; //default
}
inline cublasFillMode_t cublas_uplo_const_m (char uplo) {
	if (uplo == 'L' || uplo == 'l') return CUBLAS_FILL_MODE_LOWER;
	if (uplo == 'U' || uplo == 'u') return CUBLAS_FILL_MODE_UPPER;
	return CUBLAS_FILL_MODE_LOWER; //default
}
inline cublasDiagType_t cublas_diag_const_m (char diag) {
	if (diag == 'N' || diag == 'n') return CUBLAS_DIAG_NON_UNIT;
	if (diag == 'U' || diag == 'u') return CUBLAS_DIAG_UNIT;
	return CUBLAS_DIAG_NON_UNIT; //default
}
inline cublasSideMode_t cublas_side_const_m (char side) {
	if (side == 'L' || side == 'l') return CUBLAS_SIDE_LEFT;
	if (side == 'R' || side == 'r') return CUBLAS_SIDE_RIGHT;
	return CUBLAS_SIDE_LEFT; //default
}

// ----- CUSPARSE const value function ---------------------------------------------------------------
#if defined (CSRMV)
inline cusparseOperation_t cusparse_trans_const_m (char tran) {
	if (tran == 'N' || tran == 'n') return CUSPARSE_OPERATION_NON_TRANSPOSE;
	if (tran == 'T' || tran == 't') return CUSPARSE_OPERATION_TRANSPOSE;
	if (tran == 'C' || tran == 'c') return CUSPARSE_OPERATION_CONJUGATE_TRANSPOSE;
	return CUSPARSE_OPERATION_NON_TRANSPOSE; //default
}
#endif

// ----- CBLAS const value function ---------------------------------------------------------------
#if defined (CHECK_CBLAS)
inline CBLAS_TRANSPOSE cblas_trans_const_m (char tran) {
	if (tran == 'N' || tran == 'n') return CblasNoTrans;
	if (tran == 'T' || tran == 't') return CblasTrans;
	if (tran == 'C' || tran == 'c') return CblasConjTrans;
	return CblasNoTrans; //default
}
inline CBLAS_UPLO cblas_uplo_const_m (char uplo) {
	if (uplo == 'L' || uplo == 'l') return CblasLower;
	if (uplo == 'U' || uplo == 'u') return CblasUpper;
	return CblasLower; //default
}
inline CBLAS_DIAG cblas_diag_const_m (char diag) {
	if (diag == 'N' || diag == 'n') return CblasNonUnit;
	if (diag == 'U' || diag == 'u') return CblasUnit;
	return CblasNonUnit; //default
}
inline CBLAS_SIDE cblas_side_const_m (char side) {
	if (side == 'L' || side == 'l') return CblasLeft;
	if (side == 'R' || side == 'r') return CblasRight;
	return CblasLeft; //default
}
#endif

// ----- MBLAS compatible ---------------------------------------------------------------
#if defined (MUBLAS) && defined (CHECK_MBLAS)
inline cuddreal make_cuddreal (const qd_real v) {
	return make_cuddreal (v.x[0], v.x[1]);
}

inline cuddcomplex make_cuddcomplex (const qd_real r, const qd_real i) {
	return make_cuddcomplex (make_cuddreal (r), make_cuddreal (i));
}
#endif

#if defined (MUBLAS)
void set_mublas_handler (testingHandle_t *th, mublasHandle_t *mh) {
	mh->TxStatic = th->TxStatic;
	mh->TyStatic = th->TyStatic;
	mh->kernelnumber = th->kernelnumber;
}
#endif

void get_routine_theoretial_performance (testingHandle_t *th) {
	size_t s;
	s = th->word_size;
	size_t n;
	n = th->dim_n;
#if defined (DIM2) || defined (DIM3)
	size_t m;
	m = th->dim_m;
#endif
#if defined (DIM3)
	size_t k;
	k = th->dim_k;
#endif
#if defined (DIM1)
	th->routine_flops = (double) ROUTINE_FLOPS (n);
	th->routine_bytes = (double) ROUTINE_BYTES (n, s);
#elif defined (DIM2) 
	th->routine_flops = (double) ROUTINE_FLOPS (m, n);
	th->routine_bytes = (double) ROUTINE_BYTES (m, n, s);
#elif defined (DIM3) 
	th->routine_flops = (double) ROUTINE_FLOPS (m, n, k);
	th->routine_bytes = (double) ROUTINE_BYTES (m, n, k, s);
#endif
}

void dim_initialize (testingHandle_t *th) {
	if (th->const_m == 0) th->dim_m = th->dim_stop;
	if (th->const_n == 0) th->dim_n = th->dim_stop;
	if (th->const_k == 0) th->dim_k = th->dim_stop;
#if defined (DIM1)
	if (th->const_n == 1) {
		th->dim_start = th->dim_n;
		th->dim_stop = th->dim_n;
		th->dim_step = 1;
		th->dim_m = th->dim_n;
	}
	if (th->const_m == 1) {
		th->dim_start = th->dim_m;
		th->dim_stop = th->dim_m;
		th->dim_step = 1;
		th->dim_n = th->dim_m;
	}
#endif
}

void dim_start (testingHandle_t *th) {
	if (th->const_m == 0) th->dim_m = th->dim_start;
	if (th->const_n == 0) th->dim_n = th->dim_start;
	if (th->const_k == 0) th->dim_k = th->dim_start;
}

void dim_step (testingHandle_t *th) {
	if (th->const_m == 0) {
		if (th->mode_dim_step == 2) th->dim_m *= 2; // log mode
		else th->dim_m += th->dim_step;
	}
	if (th->const_n == 0) {
		if (th->mode_dim_step == 2) th->dim_n *= 2; // log mode
		else th->dim_n += th->dim_step;
	}
	if (th->const_k == 0) {
		if (th->mode_dim_step == 2) th->dim_k *= 2; // log mode
		else th->dim_k += th->dim_step;
	}
}

void testingCreate (int32_t argc, char **argv, testingHandle_t *th) {
 	// default ---------------------------
	th->devno = 0;
	th->mode = 'p';
	// options
	th->tranA = 'N';
	th->tranB = 'N';
	th->uplo = 'U';
	th->diag = 'N';
	th->side = 'L';

	th->TxStatic = 0;
	th->TyStatic = 0;
	th->kernelnumber = 0;
	th->cublasatomic = 0;
	th->nodisp = 0;

	th->dim_start	= 128;
	th->dim_stop	= 1024;
	th->dim_step	= 128;
	th->dim_m = 0;
	th->dim_n = 0;
	th->dim_k = 0;
	th->const_m = 0;
	th->const_n = 0;
	th->const_k = 0;
	th->incx = 1;
	th->incy = 1;
	th->test_points = 0;
	th->mode_dim_start = 0;
	th->mode_dim_stop = 0;
	th->mode_dim_step = 0;
	th->word_size = sizeof(GPU_TYPE);

 	// args ---------------------------
	int32_t opt, opt_index;
	optind = 1;
	struct option long_options[] = {
		{"devno", required_argument, NULL, 'g'},
		{"m", required_argument, NULL, 'm'},
		{"n", required_argument, NULL, 'n'},
		{"k", required_argument, NULL, 'k'},
		{"range", required_argument, NULL, 'r'},
		{"trans", required_argument, NULL, 'a'},
		{"tranb", required_argument, NULL, 'b'},
		{"uplo", required_argument, NULL, 'u'},
		{"diag", required_argument, NULL, 'd'},
		{"side", required_argument, NULL, 's'},
		{"mode", required_argument, NULL, 'o'},
		{"kernel", required_argument, NULL, 'v'},
		{"atomic", no_argument, NULL, 't'},
		{"tx", required_argument, NULL, 'x'},
		{"ty", required_argument, NULL, 'y'},
		{"points", required_argument, NULL, 'p'},
		{"help", no_argument, NULL, 'h'},
		{"incx", required_argument, NULL, 'i'},
		{"incy", required_argument, NULL, 'j'},
		{"nodisp", required_argument, NULL, 'w'},
		{0, 0, 0, 0}
	};
	while ((opt = getopt_long (argc, argv, "gmnkrabudsovtxyrphijw", long_options, &opt_index)) != -1){
		switch (opt) {
			// dim
			case 'g':{
				th->devno = atoi(optarg);
				break;
			} case 'm':{
				th->dim_m = (size_t)atoi(optarg);
				th->const_m = 1;
				break;
			} case 'n':{
				th->dim_n = (size_t)atoi(optarg);
				th->const_n = 1;
				break;
			} case 'k':{
				th->dim_k = (size_t)atoi(optarg);
				th->const_k = 1;
				break;
			} case 'r':{
				char *str, *c_start, *c_stop, *c_step;
				char par[] = ":";
				str = optarg;
				c_start = strtok (str, par);
				c_stop  = strtok (NULL, par);
				c_step	= strtok (NULL, par);

				if (strcmp (c_start, "max") == 0)
					th->mode_dim_start	= 1;
				else 
					th->dim_start	= atoi (c_start);

				if (strcmp (c_stop, "max") == 0)
					th->mode_dim_stop	= 1;
				else 
					th->dim_stop	= atoi (c_stop);

				if (strcmp (c_step, "log") == 0)
					th->mode_dim_step	= 2;
				else 
					th->dim_step	= atoi (c_step);

				break;
			} case 'a':{
			// options
				// N, T, C
				th->tranA = optarg[0];
				break;
			} case 'b':{
				// N, T, C
				th->tranB = optarg[0];
				break;
			} case 'u':{
				// U, L
				th->uplo = optarg[0];
				break;
			} case 'd':{
				// N, U
				th->diag = optarg[0];
				break;
			} case 's':{
				// L, R
				th->side = optarg[0];
				break;
			} case 'o':{
			// testing
				th->mode = optarg[0];
				break;
			} case 'v':{
				th->kernelnumber = atoi(optarg);
				break;
			} case 't':{
				th->cublasatomic = 1;
				break;
			} case 'x':{
				th->TxStatic = atoi(optarg);
				break;
			} case 'y':{
				th->TyStatic = atoi(optarg);
				break;
			} case 'p':{
				th->test_points = atoi(optarg);
				break;
			} case 'h':{
				printf ("testing program:\n");
				break;
			} case 'i':{
				th->incx = atoi(optarg);
				break;
			} case 'j':{
				th->incy = atoi(optarg);
				break;
			} case 'w':{
				th->nodisp = atoi(optarg);
				break;
			}
		}
	}

// ----- init devinfo -----
	int32_t driverVersion = 0;
	int32_t runtimeVersion = 0;
	cudaDeviceProp devprop;  
	cudaDriverGetVersion(&driverVersion);
	cudaRuntimeGetVersion(&runtimeVersion);
	cudaSetDevice (th->devno);
	cudaGetDeviceProperties (&devprop, th->devno);

	th->devprop = devprop;
	th->driverVersion = driverVersion;
	th->runtimeVersion = runtimeVersion;
	th->cc = devprop.major * 100 + devprop.minor * 10;

	switch (th->cc) {
		case 100:
			th->Cores = 8;
			th->DPCores = 1;
			break;
		case 110:
			th->Cores = 8;
			th->DPCores = 1;
			break;
		case 120:
			th->Cores = 8;
			th->DPCores = 1;
			break;
		case 130:
			th->Cores = 8;
			th->DPCores = 1;
			break;
		case 200:
			th->Cores = 32;
			th->DPCores = 16;
			break;
		case 210:
			th->Cores = 48;
			th->DPCores = 0; // ?
			break;
		case 300:
			th->Cores = 192;
			th->DPCores = 64;
			break;
		case 320:
			th->Cores = 192;
			th->DPCores = 64;
			break;
		case 350:
			th->Cores = 192;
			th->DPCores = 64;
			break;
		case 370:
			th->Cores = 192;
			th->DPCores = 64;
			break;
		case 500:
			th->Cores = 128;
			th->DPCores = 0;
			break;
		case 520:
			th->Cores = 128;
			th->DPCores = 4;
			break;
		default: 
			printf ("error: unknown architecture\n");
			break;
	}

	th->device_bytes_per_sec = (double)th->devprop.memoryClockRate * 2. * 1000. * (double)th->devprop.memoryBusWidth / 8.;
	th->device_flops_per_sec_sp = (double)th->devprop.clockRate * 1000. * 2. * (double)th->devprop.multiProcessorCount * th->Cores;
	th->device_flops_per_sec_dp = (double)th->devprop.clockRate * 1000. * 2. * (double)th->devprop.multiProcessorCount * th->DPCores;
#if defined (PREC_S) || defined (PREC_C) // single precision
	th->routine_flops_per_sec_theoretical_peak = th->device_flops_per_sec_sp;
#elif defined (PREC_D) || defined (PREC_Z) // double precision
	th->routine_flops_per_sec_theoretical_peak = th->device_flops_per_sec_dp;
#elif defined (PREC_B) || defined (PREC_DF) || defined (PREC_V) // DF-operation
	th->routine_flops_per_sec_theoretical_peak = th->device_flops_per_sec_sp / (double)DFMADCOUNT;
#elif defined (PREC_W) || defined (PREC_U) // DD-operation
	th->routine_flops_per_sec_theoretical_peak = th->device_flops_per_sec_dp / (double)DDMADCOUNT;
#else // unknown
	th->routine_flops_per_sec_theoretical_peak = 0.;
#endif

// ----- problem setting -----
	size_t max_size = 0;
	#if defined (DIM2) || defined (DIM3)
	size_t dim_1;
	#endif
	#if defined (DIM3)
	size_t dim_2;
	#endif

	switch (th->const_n + th->const_m + th->const_k) {
		case 0: {
			max_size = ROUTINE_MAXSIZE (th->devprop.totalGlobalMem, MEMMARGIN, th->word_size);
			break;
		}
		#if defined (DIM2) || defined (DIM3)
		case 1: {
			dim_1 = th->dim_m + th->dim_n + th->dim_k;
			max_size = ROUTINE_MAXSIZE_1 (th->devprop.totalGlobalMem, MEMMARGIN, th->word_size, dim_1);
			break;
		}
		#endif
		#if defined (DIM3)
		case 2: {
			if (th->const_m == 0) {
				dim_1 = th->dim_n;
				dim_2 = th->dim_k;
			} else if (th->const_n == 0) {
				dim_1 = th->dim_m;
				dim_2 = th->dim_k;
			} else if (th->const_k == 0) {
				dim_1 = th->dim_n;
				dim_2 = th->dim_m;
			}
			max_size = ROUTINE_MAXSIZE_2 (th->devprop.totalGlobalMem, MEMMARGIN, th->word_size, dim_1, dim_2);
			break;
		}
		#endif
	}

	if (th->mode_dim_start == 1) th->dim_start = max_size;
	if (th->mode_dim_stop  == 1) th->dim_stop = max_size;
	if (th->test_points != 0) {
		size_t ccc = 128, ttt = 1;
		ttt = (th->dim_stop/(th->test_points)/ccc)*ccc;
		while (ttt < 1 && ccc > 0) {
			ccc /= 2;
			ttt = (th->dim_stop/(th->test_points)/ccc)*ccc;
		}
		if (ttt < 1) ttt = 1;
		th->dim_step = ttt;
		th->dim_start = th->dim_step;
	}
	
	th->test_points = (th->dim_stop-th->dim_start)/th->dim_step;
// ---------------------
}

int32_t get_num_iter (testingHandle_t *th) {
	return std::min ((double)MAXLOOP, std::max ((double)MINLOOP, MINSEC / std::max
	(th->routine_flops / th->routine_flops_per_sec_theoretical_peak, th->routine_bytes / th->device_bytes_per_sec)));
}

double gettime () {
	struct timeval tv;
	cudaDeviceSynchronize ();
	gettimeofday (&tv, NULL);
	return tv.tv_sec + (double) tv.tv_usec * 1e-6;
}

// DF-operation
#if defined (MUBLAS)
#define SPLITTER 536870913.0 // 2^29+1
cudfreal double_to_cudfreal (const double d) {
	double t, h, l;
	t = SPLITTER * d;
	h = t - (t - d);
	l = d - h;
	return make_cudfreal ((float)h, (float)l);
}

double cudfreal_to_double (const cudfreal df) {
	return (double)df.x + (double)df.y;
}
#endif

#if (defined (CHECK_MBLAS)) && (defined (PREC_W) || defined (PREC_U))
#define RAND qdrand()/((REFERENCE_BASETYPE)RAND_MAX+1.)
#else
#define RAND rand()/((REFERENCE_BASETYPE)RAND_MAX+1.)
#endif

void init_vector (size_t n, GPU_TYPE *x) {
	int32_t sig = 1;
	for (size_t i = 0; i < n; i++) {
		if (rand()%2 == 0) sig *= -1;
		REFERENCE_BASETYPE r1 = sig*RAND;
#if defined (PREC_C) || defined (PREC_Z) || defined (PREC_U) || defined (PREC_V)
		if (rand()%2 == 0) sig *= -1;
		REFERENCE_BASETYPE r2 = sig*RAND;
		#if defined (PREC_U) || !defined (CHECK_MPACK)
		x[i] = MAKE_GPU_TYPE (r1, r2);
		#else
		x[i] = MAKE_GPU_TYPE (r1.x[0], r2.x[0]);
		#endif
#else
		#if defined (PREC_W) || !defined (CHECK_MPACK)
		x[i] = MAKE_GPU_TYPE (r1);
		#else
		x[i] = MAKE_GPU_TYPE (r1.x[0]);
		#endif
#endif
	}
}

void init_vector_1 (size_t n, GPU_TYPE *x) {
	for (size_t i = 0; i < n; i++) {
		REFERENCE_BASETYPE r1 = 1.;
#if defined (PREC_C) || defined (PREC_Z) || defined (PREC_U) || defined (PREC_V)
		REFERENCE_BASETYPE r2 = 0.;
		#if defined (PREC_U) || !defined (CHECK_MPACK)
		x[i] = MAKE_GPU_TYPE (r1, r2);
		#else
		x[i] = MAKE_GPU_TYPE (r1.x[0], r2.x[0]);
		#endif
#else
		#if defined (PREC_W) || !defined (CHECK_MPACK)
		x[i] = MAKE_GPU_TYPE (r1);
		#else
		x[i] = MAKE_GPU_TYPE (r1.x[0]);
		#endif
#endif
	}
}

#if defined (CHECK_MBLAS)
void copy_vector (size_t n, GPU_TYPE *src, MBLAS_TYPE *dst) {
	for (size_t i = 0; i < n; i++) {
#if defined (PREC_S) || defined (PREC_D) || defined (PREC_DB)
		dst[i] = qd_real ((double)src[i], 0., 0., 0.);
#endif
#if defined (PREC_W)
		dst[i] = qd_real (src[i].x, src[i].y, 0., 0.);
#endif
#if defined (PREC_C) || defined (PREC_Z) 
		dst[i] = qd_complex (qd_real ((double)src[i].x, 0., 0., 0.), qd_real ((double)src[i].y, 0., 0., 0.));
#endif
#if defined (PREC_U) 
		dst[i] = qd_complex (qd_real (src[i].x.x, src[i].x.y, 0., 0.), qd_real (src[i].y.x, src[i].y.y, 0., 0.));
#endif
#if defined (PREC_B)
		dst[i] = qd_real (cudfreal_to_double (src[i]), 0., 0., 0.);
#endif
#if defined (PREC_V) 
		dst[i] = qd_complex (qd_real (cudfreal_to_double (src[i].x), 0., 0., 0.), qd_real (cudfreal_to_double (src[i].y), 0., 0., 0.));
#endif
	}
}
#elif defined (CHECK_CBLAS)
void copy_vector (size_t n, GPU_TYPE *src, CBLAS_TYPE *dst) {
	for (size_t i = 0; i < n; i++) {
#if defined (PREC_S) || defined (PREC_D) || defined (PREC_DB)
		dst[i] = src[i];
#endif
#if defined (PREC_W)
		dst[i] = src[i].x;
#endif
#if defined (PREC_C) || defined (PREC_Z) 
		dst[i].r = src[i].x;
		dst[i].i = src[i].y;
#endif
#if defined (PREC_U) 
		dst[i].r = src[i].x.x;
		dst[i].i = src[i].y.x;
#endif
#if defined (PREC_B)
		dst[i] = cudfreal_to_double (src[i]);
#endif
#if defined (PREC_V)
		dst[i].r = cudfreal_to_double (src[i].x);
		dst[i].i = cudfreal_to_double (src[i].y);
#endif
	}
}
#elif defined (CHECK_CUBLAS)
void copy_vector (size_t n, GPU_TYPE *src, CUBLAS_TYPE *dst) {
	for (size_t i = 0; i < n; i++) {
#if defined (PREC_S) || defined (PREC_D) || defined (PREC_DB)
		dst[i] = src[i];
#endif
#if defined (PREC_W)
		dst[i] = src[i].x;
#endif
#if defined (PREC_C) || defined (PREC_Z) 
		dst[i].x = src[i].x;
		dst[i].y = src[i].y;
#endif
#if defined (PREC_U) 
		dst[i].x = src[i].x.x;
		dst[i].y = src[i].y.x;
#endif
#if defined (PREC_B)
		dst[i] = cudfreal_to_double (src[i]);
#endif
#if defined (PREC_V)
		dst[i].x = cudfreal_to_double (src[i].x);
		dst[i].y = cudfreal_to_double (src[i].y);
#endif
	}
}
#endif


// check routine (|trg-ref|/|ref|) -------------------------------------------------
void check_result (size_t m, GPU_TYPE *target, REFERENCE_TYPE *ref) {
	size_t size = sizeof (REFERENCE_TYPE);
	REFERENCE_TYPE negone = REFERENCE_NEGONE;
	REFERENCE_TYPE *diff;
	REFERENCE_BASETYPE Tnrm, Rnrm, rerr;
	diff = (REFERENCE_TYPE *) malloc (size * m);
	copy_vector (m, target, diff);

// -----------------
#if defined (CHECK_CUBLAS)
	cublasHandle_t ch;
	cublasCreate (&ch);
	REFERENCE_TYPE *dev_t, *dev_r;
	cudaMalloc ((void **) &dev_t, size * m);
	cudaMalloc ((void **) &dev_r, size * m);
	cublasSetVector (m, size, ref, 1, dev_r, 1);
	cublasSetVector (m, size, diff, 1, dev_t, 1);
	REFERENCE_AXPY (ch, m, &negone, dev_r, 1, dev_t, 1);
	cublasGetVector (m, size, dev_t, 1, diff, 1);
	REFERENCE_NRM (ch, m, dev_r, 1, &Rnrm);
	REFERENCE_NRM (ch, m, dev_t, 1, &Tnrm);
	cudaFree (dev_r);
	cudaFree (dev_t);
#else
#if !defined (CHECK_MPACK) && ( defined (PREC_C) || defined (PREC_Z) || defined (PREC_U) || defined (PREC_V) )
	REFERENCE_AXPY (m, &negone, ref, 1, diff, 1);
#else
	REFERENCE_AXPY (m, negone, ref, 1, diff, 1);
#endif
	Rnrm = REFERENCE_NRM (m, ref, 1);
	Tnrm = REFERENCE_NRM (m, diff, 1);
#endif
// -----------------

	rerr = Tnrm / Rnrm;
#if defined (CHECK_CUBLAS)
	printf("\t%1.3e\t", rerr);
#elif defined (CHECK_CBLAS)
	printf("\t%1.3e\t", rerr);
#elif defined (CHECK_MBLAS)
	printf("\t%1.3e\t(lo:\t%1.3e)\t", rerr.x[0], rerr.x[1]);
#endif
#if DEBUG_PRINT
	printf("\n%%--- result (debug-mode) ---\n");
#if defined (PREC_S) || defined (PREC_D)
	printf("%%index\tdiff\treference\ttarget\n");
	for (size_t i=0; i < m; i++)
		printf ("%zd\t%+1.2e\t%+1.2e\t%+1.2e\n", i, diff[i], ref[i], target[i]);
#endif
#if defined (PREC_C) || defined (PREC_Z)
	printf("%%index\tdiff(r,i)\treference(r,i)\ttarget(r,i)\n");
	for (size_t i=0; i < m; i++)
		#if defined (CHECK_CBLAS)
		printf ("%zd\t%+1.2e\t%+1.2e\t%+1.2e\t%+1.2e\t%+1.2e\t%+1.2e\n", (size_t)i, diff[i].r, diff[i].i, ref[i].r, ref[i].i, target[i].x, target[i].y);
		#elif defined (CHECK_CBLAS)
		printf ("%zd\t%+1.2e\t%+1.2e\t%+1.2e\t%+1.2e\t%+1.2e\t%+1.2e\n", (size_t)i, diff[i].x, diff[i].y, ref[i].x, ref[i].y, target[i].x, target[i].y);
		#endif
#endif
#endif
	free (diff);
}

// print routine -------------------------------------------------
void print_routine_name () {
	#if defined (PREC_S)
	printf ("S");
	#elif defined (PREC_D)
	printf ("D");
	#elif defined (PREC_C)
	printf ("C");
	#elif defined (PREC_Z)
	printf ("Z");
	#elif defined (PREC_W)
	printf ("W");
	#elif defined (PREC_U)
	printf ("U");
	#elif defined (PREC_B)
	printf ("B");
	#elif defined (PREC_V)
	printf ("V");
	#else
	printf ("unknown-precision,"
	#endif
	#if defined (GEMV)
	printf ("GEMV\n");
	#elif defined (SYMV)
	printf ("SYMV\n");
	#elif defined (HEMV)
	printf ("HEMV\n");
	#elif defined (TRMV)
	printf ("TRMV\n");
	#elif defined (TRSV)
	printf ("TRSV\n");
	#elif defined (AXPY)
	printf ("AXPY\n");
	#elif defined (COPY)
	printf ("COPY\n");
	#elif defined (DOT)
	printf ("DOT\n");
	#elif defined (DOTU)
	printf ("DOTU\n");
	#elif defined (GER)
	printf ("GER\n");
	#elif defined (CSRMV)
	printf ("CSRMV\n");
	#elif defined (GEMM)
	printf ("GEMM\n");
	#else
	printf ("unknown-routine\n"
	#endif
}

void print_library_name () {
	#if defined (MUBLAS)
	printf ("MUBLAS\n");
	#elif defined (CUBLAS)
	printf ("CUBLAS\n");
	#elif defined (KBLAS)
	printf ("KBLAS\n");
	#elif defined (MAGMA)
	printf ("MAGMA\n");
	#elif defined (CULA)
	printf ("CULA\n");
	#elif defined (ASPEN)
	printf ("ASPEN.K2\n");
	#else
	printf ("unknown-library\n");
	#endif
}

void print_info1 (testingHandle_t *th) {
	if (!th->nodisp) {
		system("echo -n '% Testing info -----------------------------\n'");
		system("echo -n '% \tDate\t'; date");
		system("echo -n '% \tHost\t'; hostname");
		system("echo -n '% \t'");
		system("head -n 1 /proc/driver/nvidia/version");
		system("echo -n '% \t'");
		system("tail -n 1 /proc/driver/nvidia/version");
		printf("%% \tDevice\t%s\n", th->devprop.name);
		printf("%% \tTheoretical Peak Memory Bandwidth\t%3.1f [GB/s]\n", th->device_bytes_per_sec * 1.e-9);
		printf("%% \tTheoretical Peak Performance (SP)\t%4.1f [GFlops]\n", th->device_flops_per_sec_sp * 1.e-9);
		printf("%% \tTheoretical Peak Performance (DP)\t%4.1f [GFlops]\n", th->device_flops_per_sec_dp * 1.e-9);
		printf("%% \tTotal Gmem\t%4.1f [GB]\n", th->devprop.totalGlobalMem * 1e-9);
		printf("%% \tCC\t%d.%d\n", th->devprop.major, th->devprop.minor);
		printf("%% \tECC Enabled\t%d\n", th->devprop.ECCEnabled);
		printf("%% \tMultiProcessorCount\t%d\n", th->devprop.multiProcessorCount);
		printf("%% \tDriver\t%d\n", th->driverVersion);
		printf("%% \tRuntime\t%d\n", th->runtimeVersion);
		printf ("%%\tRoutine\t");
		print_routine_name ();
		printf ("%%\tLibrary\t");
		print_library_name ();
		if (th->mode == 'c') {
			printf ("%%\tCheck\t");
			#if defined (CHECK_CUBLAS)
			printf ("CUBLAS\n");
			#elif defined (CHECK_CBLAS)
			printf ("CBLAS\n");
			#elif defined (CHECK_MBLAS)
			printf ("MBLAS\n");
			#endif
		}
		printf("%% \tGmem allocation\t");
		#if PITCHALLOC
		printf("pitched\n");
		#else
		printf("non-pitched\n");
		#endif
		printf ("%%\tMINLOOP\t%d\n", MINLOOP);
		printf ("%%\tMAXLOOP\t%d\n", MAXLOOP);
		printf ("%%\tMINSEC\t%f\n", MINSEC);
		printf ("%% ------------------------------------------\n");
		printf ("%% BLAS parameter ---------------------------\n");
		#if !defined (DIM1) 
		printf ("%%\ttranA\t%c\n", th->tranA);
		#endif
		#if defined (DIM3) 
		printf ("%%\ttranB\t%c\n", th->tranB);
		#endif
		#if !defined (DIM1)
		printf ("%%\tuplo\t%c\n", th->uplo);
		printf ("%%\tdiag\t%c\n", th->diag);
		printf ("%%\tside\t%c\n", th->side);
		#endif
		printf ("%%\tdim_start\t%zd\n", (size_t)th->dim_start);
		printf ("%%\tdim_stop\t%zd\n", (size_t)th->dim_stop);
		printf ("%%\tdim_step\t%zd\n", (size_t)th->dim_step);
		#if !defined (DIM1)
		printf ("%%\tM\t%zd\n", (size_t)th->dim_m);
		#endif
		printf ("%%\tN\t%zd\n", (size_t)th->dim_n);
		#if defined (DIM3)
		printf ("%%\tK\t%zd\n", (size_t)th->dim_k);
		#endif
		#if !defined (DIM3) 
		printf ("%%\tincx\t%zd\n", (size_t)th->incx);
		#endif
		#if defined (DIM2) 
		printf ("%%\tincy\t%zd\n", (size_t)th->incy);
		#endif
		printf ("%%\tcublasatomic\t%d\n", th->cublasatomic);
		printf ("%% ------------------------------------------\n");
	}
}

void print_info2 (testingHandle_t *th) {
	if (!th->nodisp) {
		printf ("%% Evaluation result ------------------------\n");
		printf ("%% M\tN\tK");
		if (th->mode == 'p')
			printf ("\tTime[sec]\t[GFlops]\t[GB/s]\t(#iter)");
		if (th->mode == 'c')
			printf ("\tERROR");
		#if defined (MUBLAS)
		printf ("\ttx\tty\tth\tyx\tsb\tbd\tbm\tWrpOcp\tBlkOcp\tGrdOcp\tCount\tkernel#");
		#endif
		#if defined (USE_NVML)
		printf ("\tTemp.bf[C]\tTemp.af[C]");
		#endif
		printf ("\n");
	}
}

#if defined (MUBLAS)
void print_mublas_info1 (testingHandle_t *th, mublasHandle_t *mh) {
	printf ("\t%zd\t%zd\t%zd\t%1.3f\t%zd\t%zd\t%1.3f\t%1.3f\t%1.3f\t%1.3f\t%d\t%d",
	(size_t)mh->Tx, (size_t)mh->Ty, (size_t)(mh->Tx*mh->Ty), (float)mh->Ty/(float)mh->Tx,
	(size_t)mh->ShmPerBlk, (size_t)mh->BlkPerDev, (float)mh->BlkPerDev/(float)mh->devprop_multiProcessorCount,
	mh->WrpOcp, mh->BlkOcp, mh->GrdOcp, mh->count, mh->kernelnumber);
}

void print_mublas_info2 (testingHandle_t *th, mublasHandle_t *mh) {
	if (!th->nodisp) {
		printf ("%% ------------------------------------------\n");
		printf ("%% MUBLAS info (last called) ----------------\n");
		printf ("%% \tmublasVersion\t%d\n", mh->mublasVersion);
		printf ("%% \tbinaryVersion\t%d\n", mh->funcatb.binaryVersion);
		printf ("%% \tconstSizeBytes\t%zd\n", mh->funcatb.constSizeBytes);
		printf ("%% \tlocalSizeBytes\t%zd\n", mh->funcatb.localSizeBytes);
		printf ("%% \tmaxThreadsPerBlock\t%d\n", mh->funcatb.maxThreadsPerBlock);
		printf ("%% \tnumRegs\t%d\n", mh->funcatb.numRegs);
		printf ("%% \tptxVersion\t%d\n", mh->funcatb.ptxVersion);
		printf ("%% \tsharedSizeBytes\t%zd\n", mh->funcatb.sharedSizeBytes);
		printf ("%% ------------------------------------------\n");
		printf ("%% \tTxStatic\t%zd\n", (size_t)mh->TxStatic);
		printf ("%% \tTyStatic\t%zd\n", (size_t)mh->TyStatic);
		printf ("%% ------------------------------------------\n");
		printf ("%% \tThMin\t%zd\n", (size_t)mh->ThMin);
		printf ("%% \tThMax\t%zd\n", (size_t)mh->ThMax);
		printf ("%% \tTxMin\t%zd\n", (size_t)mh->TxMin);
		printf ("%% \tTxMax\t%zd\n", (size_t)mh->TxMax);
		printf ("%% \tTxStep\t%zd\n", (size_t)mh->TxStep);
		printf ("%% \tTxPrefer\t%c\n", mh->TxPrefer);
		printf ("%% \tTyPrefer\t%c\n", mh->TyPrefer);
		printf ("%% \tTyPerTxMax\t%f\n", mh->TyPerTxMax);
		printf ("%% ------------------------------------------\n");
		printf ("%% \tWrpOcpMin\t%1.4f\n", mh->WrpOcpMin);
		printf ("%% \tBlkOcpMin\t%1.4f\n", mh->BlkOcpMin);
		printf ("%% \tGrdOcpMin\t%1.4f\n", mh->GrdOcpMin);
		printf ("%% ------------------------------------------\n");
	}
}
#endif

void print_devprop (testingHandle_t *th, cudaDeviceProp *devprop) {
	if (!th->nodisp) {
		printf ("totalGlobalMem\t%zd\n", devprop->totalGlobalMem);
		printf ("sharedMemPerBlock\t%zd\n", devprop->sharedMemPerBlock);
		printf ("regsPerBlock\t%d\n", devprop->regsPerBlock);
		printf ("warpSize\t%d\n", devprop->warpSize);
		printf ("memPitch\t%zd\n", devprop->memPitch);
		printf ("maxThreadsPerBlock\t%d\n", devprop->maxThreadsPerBlock);
		printf ("maxThreadsDim[0]\t%d\n", devprop->maxThreadsDim[0]);
		printf ("maxThreadsDim[1]\t%d\n", devprop->maxThreadsDim[1]);
		printf ("maxThreadsDim[2]\t%d\n", devprop->maxThreadsDim[2]);
		printf ("maxGridSize[0]\t%d\n", devprop->maxGridSize[0]);
		printf ("maxGridSize[1]\t%d\n", devprop->maxGridSize[1]);
		printf ("maxGridSize[2]\t%d\n", devprop->maxGridSize[2]);
		printf ("clockRate\t%d\n", devprop->clockRate);
		printf ("totalConstMem\t%zd\n", devprop->totalConstMem);
		printf ("major\t%d\n", devprop->major);
		printf ("minor\t%d\n", devprop->minor);
		printf ("textureAlignment\t%zd\n", devprop->textureAlignment);
		printf ("texturePitrchAlignment\t%zd\n", devprop->texturePitchAlignment);
		printf ("deviceOverlap\t%d\n", devprop->deviceOverlap);
		printf ("multiProcessorCount\t%d\n",  devprop->multiProcessorCount);
		printf ("kernelExecTimeoutEnabled\t%d\n", devprop->kernelExecTimeoutEnabled);
		printf ("integrated\t%d\n", devprop->integrated);
		printf ("canMapHostMemory\t%d\n", devprop->canMapHostMemory);
		printf ("computeMode\t%d\n", devprop->computeMode);
		printf ("maxTexture1D\t%d\n", devprop->maxTexture1D);
		printf ("maxTexture1DMipmap\t%d\n", devprop->maxTexture1DMipmap);
		printf ("maxTexture1DLinear\t%d\n", devprop->maxTexture1DLinear);
		printf ("maxTexture2D[0]\t%d\n", devprop->maxTexture2D[0]);
		printf ("maxTexture2D[1]\t%d\n", devprop->maxTexture2D[1]);
		printf ("maxTexture2DMipmap[0]\t%d\n", devprop->maxTexture2DMipmap[0]);
		printf ("maxTexture2DMipmap[1]\t%d\n", devprop->maxTexture2DMipmap[1]);
		printf ("maxTexture2DLinear[0]\t%d\n", devprop->maxTexture2DLinear[0]);
		printf ("maxTexture2DLinear[1]\t%d\n", devprop->maxTexture2DLinear[1]);
		printf ("maxTexture2DLinear[2]\t%d\n", devprop->maxTexture2DLinear[2]);
		printf ("maxTexture2DGather[0]\t%d\n", devprop->maxTexture2DGather[0]);
		printf ("maxTexture2DGather[1]\t%d\n", devprop->maxTexture2DGather[1]);
		printf ("maxTexture3D[0]\t%d\n", devprop->maxTexture3D[0]);
		printf ("maxTexture3D[1]\t%d\n", devprop->maxTexture3D[1]);
		printf ("maxTexture3D[2]\t%d\n", devprop->maxTexture3D[2]);
		printf ("maxTexture3DAlt[0]\t%d\n", devprop->maxTexture3DAlt[0]);
		printf ("maxTexture3DAlt[1]\t%d\n", devprop->maxTexture3DAlt[1]);
		printf ("maxTexture3DAlt[2]\t%d\n", devprop->maxTexture3DAlt[2]);
		printf ("maxTextureCubemap\t%d\n", devprop->maxTextureCubemap);
		printf ("maxTexture1DLayered[0]\t%d\n", devprop->maxTexture1DLayered[0]);
		printf ("maxTexture1DLayered[1]\t%d\n", devprop->maxTexture1DLayered[1]);
		printf ("maxTexture2DLayered[0]\t%d\n", devprop->maxTexture2DLayered[0]);
		printf ("maxTexture2DLayered[1]\t%d\n", devprop->maxTexture2DLayered[1]);
		printf ("maxTexture2DLayered[2]\t%d\n", devprop->maxTexture2DLayered[2]);
		printf ("maxTextureCubemapLayered[0]\t%d\n", devprop->maxTextureCubemapLayered[0]);
		printf ("maxTextureCubemapLayered[1]\t%d\n", devprop->maxTextureCubemapLayered[1]);
		printf ("maxSurface1D\t%d\n", devprop->maxSurface1D);
		printf ("maxSurface2D[0]\t%d\n", devprop->maxSurface2D[0]);
		printf ("maxSurface2D[1]\t%d\n", devprop->maxSurface2D[1]);
		printf ("maxSurface3D[0]\t%d\n", devprop->maxSurface3D[0]);
		printf ("maxSurface3D[1]\t%d\n", devprop->maxSurface3D[1]);
		printf ("maxSurface3D[2]\t%d\n", devprop->maxSurface3D[2]);
		printf ("maxSurface1DLayered[0]\t%d\n", devprop->maxSurface1DLayered[0]);
		printf ("maxSurface1DLayered[1]\t%d\n", devprop->maxSurface1DLayered[1]);
		printf ("maxSurface2DLayered[0]\t%d\n", devprop->maxSurface2DLayered[0]);
		printf ("maxSurface2DLayered[1]\t%d\n", devprop->maxSurface2DLayered[1]);
		printf ("maxSurface2DLayered[2]\t%d\n", devprop->maxSurface2DLayered[2]);
		printf ("maxSurfaceCubemap\t%d\n", devprop->maxSurfaceCubemap);
		printf ("maxSurfaceCubemapLayered[0]\t%d\n", devprop->maxSurfaceCubemapLayered[0]);
		printf ("maxSurfaceCubemapLayered[1]\t%d\n", devprop->maxSurfaceCubemapLayered[1]);
		printf ("surfaceAlignment\t%zd\n", devprop->surfaceAlignment);
		printf ("concurrentKernels\t%d\n", devprop->concurrentKernels);
		printf ("ECCEnabled\t%d\n", devprop->ECCEnabled);
		printf ("pciBusID\t%d\n", devprop->pciBusID);
		printf ("pciDeviceID\t%d\n", devprop->pciDeviceID);
		printf ("pciDomainID\t%d\n", devprop->pciDomainID);
		printf ("tccDriver\t%d\n", devprop->tccDriver);
		printf ("asyncEngineCount\t%d\n", devprop->asyncEngineCount);
		printf ("unifiedAddressing\t%d\n", devprop->unifiedAddressing);
		printf ("memoryClockRate\t%d\n", devprop->memoryClockRate);
		printf ("memoryBusWidth\t%d\n", devprop->memoryBusWidth);
		printf ("l2CacheSize\t%d\n", devprop->l2CacheSize);
		printf ("maxThreadsPerMultiProcessor\t%d\n", devprop->maxThreadsPerMultiProcessor);
		printf ("streamPrioritiesSupported\t%d\n", devprop->streamPrioritiesSupported);
		printf ("globalL1CacheSupported\t%d\n", devprop->globalL1CacheSupported);
		printf ("localL1CacheSupported\t%d\n", devprop->localL1CacheSupported);
		printf ("sharedMemPerMultiprocessor\t%zd\n", devprop->sharedMemPerMultiprocessor);
		printf ("regsPerMultiprocessor\t%d\n", devprop->regsPerMultiprocessor);
//		printf ("managedMemSupported\t%d\n", devprop->managedMemSupported);
		printf ("isMultiGpuBoard\t%d\n", devprop->isMultiGpuBoard);
		printf ("multiGpuBoardGroupID\t%d\n", devprop->multiGpuBoardGroupID);
		printf ("----------------\n");
	}
}
